# -*- coding: utf8 -*-
# Copyright (c) 2018-2020 THL A29 Limited, a Tencent company. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import

import os
from ti_containers import constant
from ti_containers.executer import Worker
from ti_containers.utils import (
    dns_lookup,
    get_logger
)

logger = get_logger()

def run(config):
    logger.info('DNS lookups begin')
    for host in config.hosts:
        dns_lookup(host)

    os.environ['MASTER_ADDR'] = config.hosts[0]
    os.environ['MASTER_PORT'] = str(constant.PYTORCH_MASTER_PORT)
    return Worker(config)



# -*- coding: utf8 -*-
# Copyright (c) 2018-2020 THL A29 Limited, a Tencent company. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import

import os
import subprocess
import sys
import time

from ti_containers import constant
from ti_containers import utils
from ti_containers.exception import RunUserEntrypointError
from ti_containers.utils import get_logger

logger = get_logger()

class Worker(object):
    def __init__(self, config):
        self.config = config

    def _get_command(self):
        return [sys.executable, self.config.user_entry_point] + self.config.get_args()

    '''
    def run(self):
        command = self._get_command()
        logger.info("Training command: %s", " ".join(command))

        process = subprocess.Popen(
            command, stderr=subprocess.PIPE, env=os.environ
        )

        _, stderr = process.communicate()
        exit_code = process.poll()
        if exit_code != 0:
            raise RunUserEntrypointError(command=" ".join(command), output=stderr)
    '''

    def run(self):
        command = self._get_command()
        logger.info("Training command: %s", " ".join(command))

        process = subprocess.Popen(
            command, env=os.environ
        )

        exit_code = process.wait()
        if exit_code != 0:
            raise RunUserEntrypointError(command=" ".join(command), output="")

class MPIWorker(Worker):
    def __init__(self, config):
        self.config = config

    def run(self):
        logger.info('MPI worker start')
        utils.start_sshd_process(constant.SSH_PORT)

        logger.info('Waiting orted process finish.')
        import psutil
        for t in range(int(constant.TIMEOUT_SECONDS / constant.INTERVAL_SECONDS)):
            procs = [p for p in psutil.process_iter(attrs=['name']) if p.info['name'] == "orted"]
            if procs:
                break

            time.sleep(constant.INTERVAL_SECONDS)

        if not procs:
            raise RuntimeError("ERROR. orted process can not work")

        logger.info('Orted process work.')
        psutil.wait_procs(procs)

        logger.info('MPI process finish.')
        time.sleep(30)


class MPILanucher(Worker):
    def __init__(self, config):
        self.config = config

    def _get_command(self):
        command = ['mpirun', '--allow-run-as-root', '-display-allocation', '--display-map']
        np = self.config.mpi_process_per_host * len(self.config.hosts)
        command.extend(['-np', str(np)])
        if not self.config.mpi_operator_submit:
            utils.start_sshd_process(constant.SSH_PORT)
            self._wait_workers_ready()

            host_ips = [utils.dns_lookup(host) for host in self.config.hosts]
            if self.config.mpi_process_per_host == 1:
                mpi_hosts = host_ips
            else:
                mpi_hosts = ['%s:%s' % (host, self.config.mpi_process_per_host) for host in host_ips]

            command.extend([
                '--host', ','.join(mpi_hosts)
            ])

        nccl_options, custom_options = utils.get_mpi_options(self.config.custom_mpi_options)
        command.extend([
            '-bind-to', 'None',
            '-map-by', 'slot',
            '-mca', 'btl_tcp_if_include', self.config.network_interface_name,
            '-mca', 'oob_tcp_if_include', self.config.network_interface_name,
            '-mca', 'pml', 'ob1', '-mca', 'btl', '^openib',
            '-mca', 'orte_abort_on_non_zero_status', '1',
            '-x', 'NCCL_DEBUG=%s' % nccl_options.NCCL_DEBUG,
            '-x', 'NCCL_SOCKET_IFNAME=%s' % self.config.network_interface_name,
            '-x', 'LD_LIBRARY_PATH',
            '-x', 'PATH',
        ])

        command.extend(custom_options)

        for name in self.config.get_envs():
            command.extend(['-x', name])

        command.extend([sys.executable, '-m', 'mpi4py', self.config.user_entry_point])
        command.extend(self.config.get_args())
        return command

    def _wait_workers_ready(self):
        logger.info('Wait MPI workers ssh ready')
        for host in self.config.hosts:
            if host == self.config.current_host:
                continue

            pre = int(time.time())
            while not utils.check_ssh_connect(host, constant.SSH_PORT):
                time.sleep(constant.INTERVAL_SECONDS)
                now = int(time.time())
                if now - pre > constant.TIMEOUT_SECONDS:
                    raise RuntimeError("Training failed. worker %s ssh not ready " % host)

            logger.info('Worker %s ready', host)



# -*- coding: utf8 -*-
# Copyright (c) 2018-2020 THL A29 Limited, a Tencent company. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import

import json
import logging
import os

import six
from ti_containers import constant
from ti_containers import utils


class Config(object):
    def load(self):
        # parse hyperparameters
        hps = utils.read_json_file(constant.HYPERPARAMETERS_PATH)
        hps = utils.format_hyperparameters(hps)
        extend_hps, train_hps = utils.extract_hyperparameters(hps, constant.TI_PREFIX)

        self.network_interface_name = extend_hps.get(constant.TI_NETWORK_INTERFACE_NAME, constant.DEFAULT_NIC_NAME)
        self.log_level = extend_hps.get(constant.TI_CONTAINER_LOG_LEVEL, logging.INFO)
        self.user_entry_point = extend_hps.get(constant.TI_PROGRAM, '')
        self.parameter_server_enabled = extend_hps.get(constant.TI_PARAMETER_SERVER_ENABLED, '')
        self.mpi_enabled = extend_hps.get(constant.TI_MPI_ENABLED, '')
        self.mpi_process_per_host = extend_hps.get(constant.TI_MPI_NUM_OF_PROCESSES_PER_HOST, '')
        self.custom_mpi_options = extend_hps.get(constant.TI_MPI_CUSTOM_MPI_OPTIONS, '')

        self.train_hps = train_hps

        # parse input data config
        input_data_config = utils.read_json_file(constant.INPUT_DATA_CONFIG_PATH)
        self.input_channels = {v: os.path.join(constant.INPUT_CHANNLE_PATH, v) for v in input_data_config}

        # parse resource config
        resource_config = utils.read_json_file(constant.RESOURCE_CONFIG_PATH)
        self.current_host = resource_config[constant.CURRENT_HOST]
        self.hosts = resource_config[constant.HOSTS]
        self.mpi_operator_submit = resource_config.get(constant.MPI_OPERATOR_SUBMIT, False)
        self.is_master = self.current_host == self.hosts[0]

        if self.mpi_operator_submit:
            self.is_master = True

        # parse training framework
        self.training_framework = os.environ.get(constant.TI_TRAINING_FRAMEWORK, None)

        self.model_dir = constant.MODEL_PATH
        self.input_config_dir = constant.INPUT_CONFIG_PATH
        self.output_data_dir = constant.OUTPUT_DATA_PATH

        self.update_os_envs()

    def get_args(self):
        args = []
        for k, v in self.train_hps.items():
            args.append('--' + k)
            if str(v).strip():
                args.append(str(v))

        return args

    def get_envs(self):
        envs = {
            "network_interface_name": self.network_interface_name,
            "user_entry_point": self.user_entry_point,
            "hps": self.train_hps,
            "channels": list(self.input_channels.keys()),
            "num_cpus": utils.get_cpu_nums(),
            "num_gpus": utils.get_gpu_nums(),
            "log_level": self.log_level,
            "model_dir": self.model_dir,
            "output_data_dir": self.output_data_dir,
            "input_config_dir": self.input_config_dir,
            "hosts":self.hosts,
            "current_host":self.current_host,
        }

        for name, path in self.input_channels.items():
            envs['channel_%s' % name] = path

        env_vars = {}
        for k, v in envs.items():
            k = constant.TM_PREFIX + str(k).upper()
            if isinstance(v, six.string_types):
                v = str(v)
            else:
                v = json.dumps(v)

            env_vars[k] = v

        return env_vars

    def update_os_envs(self):
        for name, value in self.get_envs().items():
            os.environ[name] = value



# -*- coding: utf8 -*-
# Copyright (c) 2018-2020 THL A29 Limited, a Tencent company. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import

import os
import json
import multiprocessing
import tensorflow as tf

from ti_containers import constant
from ti_containers import utils
from ti_containers.executer import Worker
from ti_containers.utils import get_logger

logger = get_logger()

def _build_config(config):
    master, ps, workers = config.hosts[0], config.hosts[0], config.hosts[1:]
    tf_config = {
        'cluster': {
            'master': ['{}:{}'.format(master, constant.TF_WORKER_PORT)],
            'ps': ['{}:{}'.format(ps, constant.TF_PS_PORT)],
            'worker': ['{}:{}'.format(host, constant.TF_WORKER_PORT) for host in workers],
        },
        'environment': 'cloud'
    }

    if config.is_master:
        tf_config['task'] = {'index': 0, 'type': 'master'}
    else:
        tf_config['task'] = {'index': workers.index(config.current_host), 'type': 'worker'}

    return tf_config


def _run_ps_server(tf_config):
    cluster_spec = tf.train.ClusterSpec(tf_config['cluster'])
    if tf.__version__ < '2.0.0':
        config = tf.ConfigProto(device_count={'GPU': 0})
        server = tf.train.Server(cluster_spec, job_name='ps', task_index=0, config=config)
    else:
        config = tf.compat.v1.ConfigProto(device_count={'GPU': 0})
        server = tf.distribute.Server(cluster_spec, job_name='ps', task_index=0, config=config)

    multiprocessing.Process(target=lambda: server.join()).start()


def run(config):
    logger.info('DNS lookups begin')
    for host in config.hosts:
        utils.dns_lookup(host)

    if config.parameter_server_enabled and len(config.hosts) > 1:
        tf_config = _build_config(config)
        os.environ["TF_CONFIG"] = json.dumps(tf_config)
        if config.is_master:
            _run_ps_server(tf_config)

    return Worker(config)



# -*- coding: utf8 -*-
# Copyright (c) 2018-2020 THL A29 Limited, a Tencent company. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import

import os

TM_PREFIX = 'TM_'
TI_PREFIX = 'ti_'

ROOT_BASE_PATH = "/opt/ml/"
CODE_PATH = os.path.join(ROOT_BASE_PATH, "code")
MODEL_PATH = os.path.join(ROOT_BASE_PATH, "model")
INPUT_PATH = os.path.join(ROOT_BASE_PATH, "input")
OUTPUT_PATH = os.path.join(ROOT_BASE_PATH, "output")

INPUT_CHANNLE_PATH = os.path.join(INPUT_PATH, "data")
INPUT_CONFIG_PATH = os.path.join(INPUT_PATH, "config")

OUTPUT_SUCCESS_PATH = os.path.join(OUTPUT_PATH, "success")
OUTPUT_FAILURE_PATH = os.path.join(OUTPUT_PATH, "failure")
OUTPUT_DATA_PATH = os.path.join(OUTPUT_PATH, "data")

HYPERPARAMETERS_PATH = os.path.join(INPUT_CONFIG_PATH, "hyperparameters.json")
RESOURCE_CONFIG_PATH = os.path.join(INPUT_CONFIG_PATH, "resourceconfig.json")
INPUT_DATA_CONFIG_PATH = os.path.join(INPUT_CONFIG_PATH, "inputdataconfig.json")

TI_REGION = "ti_region"
TI_JOB_NAME = "ti_job_name"
TI_NETWORK_INTERFACE_NAME = "ti_network_interface_name"
TI_CONTAINER_LOG_LEVEL = "ti_container_log_level"
TI_PROGRAM = "ti_program"
TI_SUBMIT_DIRECTORY = "ti_submit_directory"
TI_PARAMETER_SERVER_ENABLED = "ti_parameter_server_enabled"
TI_MPI_ENABLED = "ti_mpi_enabled"
TI_MPI_NUM_OF_PROCESSES_PER_HOST = "ti_mpi_num_of_processes_per_host"
TI_MPI_CUSTOM_MPI_OPTIONS = "ti_mpi_custom_mpi_options"
TI_TRAINING_FRAMEWORK = 'TI_TRAINING_FRAMEWORK'

CURRENT_HOST = "current_host"
HOSTS = "hosts"

TF_WORKER_PORT = 5555
TF_PS_PORT = 5556

PYTORCH_MASTER_PORT = 6666

MXNET_PS_PORT = 7777
MXNET_PS_VERBOSE = 0

TIMEOUT_SECONDS = 300
INTERVAL_SECONDS = 5

DEFAULT_NIC_NAME = 'eth0'

SSH_PORT = 23456

MPI_OPERATOR_SUBMIT = "mpi_operator_submit"




# -*- coding: utf8 -*-
# Copyright (c) 2018-2020 THL A29 Limited, a Tencent company. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import

import traceback
import os
from importlib import import_module

from ti_containers import constant
from ti_containers import utils
from ti_containers.config import Config
from ti_containers.exception import RunUserEntrypointError
from ti_containers.executer import (
    Worker,
    MPILanucher,
    MPIWorker,
)
from ti_containers.utils import get_logger, configure_logger

logger = get_logger()


def framework_training(config):
    framework = config.training_framework
    logger.info('Training framework %s load', framework)

    module = "ti_containers." + framework.lower() + ".prepare"
    worker = getattr(import_module(module), "run")
    worker(config).run()


def common_training(config):
    logger.info('Training entry point load')

    if config.mpi_enabled:
        if config.is_master:
            executor = MPILanucher(config)
        else:
            executor = MPIWorker(config)
    else:
        executor = Worker(config)

    executor.run()


def main():
    try:
        # config load
        config = Config()
        config.load()

        # set log level
        configure_logger(config.log_level)

        # train start
        if config.training_framework and not config.mpi_enabled:
            framework_training(config)
        else:
            common_training(config)

        logger.info("Training success")
        utils.write_file(constant.OUTPUT_SUCCESS_PATH, "success")
    except RunUserEntrypointError as e:
        error_msg = "Run error: \n%s\n" % (str(e))
        logger.error(error_msg)
        utils.write_file(constant.OUTPUT_FAILURE_PATH, error_msg)
        os._exit(1)
    except Exception as e:
        error_msg = "Training error: \n%s\n" % (str(e))
        logger.error(error_msg)
        utils.write_file(constant.OUTPUT_FAILURE_PATH, error_msg)
        os._exit(1)



# -*- coding: utf8 -*-

from __future__ import absolute_import

from setuptools import setup, find_packages


def read(fname):
    with open(fname, 'r') as file:
        return file.read()


setup(
    name="ti_containers",
    version=read("version").strip(),
    url='https://cloud.tencent.com',
    license="Apache License 2.0",
    description="Open source library for training container on TencentCloud TIONE.",
    long_description=read("README.md"),
    packages=find_packages("src"),
    package_dir={"": "src"},
    install_requires=['retrying==1.3.3', 'six==1.12.0'],
    long_description_content_type='text/markdown',
    entry_points={
        'console_scripts': ['train=ti_containers.train:main'],
    },
    author="TencentCloud TIONE",
    keywords="TencentCloud ML TI AI Training",
)



# -*- coding: utf8 -*-
# Copyright (c) 2018-2020 THL A29 Limited, a Tencent company. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

class RunUserEntrypointError(Exception):
    def __init__(self, command, output):
        self.command = command
        self.output = output
        super(RunUserEntrypointError, self).__init__()

    def __str__(self):
        error_msg = "\n%s" % self.output if self.output else ""
        message = '%s:\nCommand "%s"%s' % (type(self).__name__, self.command, error_msg)
        return message.strip()



# -*- coding: utf8 -*-
# Copyright (c) 2018-2020 THL A29 Limited, a Tencent company. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import

import argparse
import json
import logging
import multiprocessing
import os
import shlex
import socket
import subprocess

from retrying import retry

def get_logger():
    return logging.getLogger("ti")

def configure_logger(level):
    logging.basicConfig(format='%(asctime)s %(name)-4s %(levelname)-4s %(message)s', level=int(level))

def file_exist(path):
    if not os.path.isfile(path):
        return False

    return True


def read_json_file(path):
    with open(path, 'r') as f:
        return json.load(f)


def write_json_file(obj, path):
    with open(path, 'w') as f:
        json.dump(obj, f)


def write_file(path, data):
    with open(path, 'w') as f:
        f.write(data)


def format_hyperparameters(hps):
    hyperparameters = {}
    for k, v in hps.items():
        try:
            v = json.loads(v)
        except (ValueError, TypeError):
            pass

        hyperparameters[k] = v

    return hyperparameters


def extract_hyperparameters(hps, prefix):
    extend_hps = {}
    train_hps = {}

    for k, v in hps.items():
        if k.startswith(prefix):
            extend_hps[k] = v
        else:
            train_hps[k] = v

    return extend_hps, train_hps


def get_gpu_nums():
    try:
        cmd = shlex.split('nvidia-smi --list-gpus')
        output = subprocess.check_output(cmd).decode('utf-8')
        return sum([1 for x in output.split('\n') if x.startswith('GPU ')])
    except Exception as e:
        return 0


def get_cpu_nums():
    return multiprocessing.cpu_count()


def get_mpi_options(mpi_options):
    mpi_parser = argparse.ArgumentParser()
    mpi_parser.add_argument("--NCCL_DEBUG", default="INFO", type=str)
    return mpi_parser.parse_known_args(mpi_options.split())


def start_sshd_process(port):
    subprocess.Popen(["/usr/sbin/sshd", "-D", "-p", str(port)])


def check_ssh_connect(host, port):
    import paramiko

    try:
        client = paramiko.SSHClient()
        client.load_system_host_keys()
        client.set_missing_host_key_policy(paramiko.AutoAddPolicy())
        client.connect(host, port=port)
        client.close()
        return True
    except Exception as e:
        get_logger().info('Cannot connect host %s, %s', host, e)
        return False


@retry(stop_max_delay=1000 * 60 * 10,
       wait_exponential_multiplier=200,
       wait_exponential_max=50000)
def dns_lookup(host):
    return socket.gethostbyname(host)
